﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

namespace ClassLibraryGeometry
{
    /// <summary>
    /// Структура PointD определяет точку на плоскости,
    /// заданную декартовыми и полярными координатами.
    /// Над точкой определены операции перемещения по горизонали и вертикали,
    /// перемещение по радиусу и вращение по окружности
    /// </summary>
    public struct PointD
    {
        //константы класса

        const double EPS = 1.0E-7;
        const string ERR_MESSAGE1 =
            "Ошибка: r < 0: !!! Замена r на |r| !!!  ";

        //поля класса 

        /// <summary>
        /// x, y - декартовы координаты
        /// </summary>
        double x, y;
        /// <summary>
        /// r, fi - полярные координаты
        /// </summary>
        double r, fi;
        //конструкторы класса

        /// <summary>
        /// Декартов конструктор
        /// </summary>
        /// <param name="x">горизонтальная координата по оси x</param>
        /// <param name="y">вертикальная координата по оси y</param>
        public PointD(double x, double y)
        {
            this.x = x; this.y = y;
            r = Math.Sqrt(x * x + y * y);
            r = Math.Sqrt(x * x + y * y);
            fi = Math.Atan2(y, x);
        }
        /// <summary>
        /// Конструктор, которому передаются полярные координаты точки - r, fi
        /// радиус r должен быть больше или равен нуля.
        /// Отрицательный радиус заменяется его значением по модулю!
        /// </summary>
        /// <param name="r">расстояние до начала координат</param>
        /// <param name="fi">угол в радианах, 
        ///            отсчитываемый от оси х против часовой стрелки</param>
        /// <param name="tip">тип конструктора, значение не важно, 
        ///            обычно слово polar</param>
        public PointD(double r, double fi, string tip)
        {
            try
            {
                if (r < 0) throw (new PointDException(ERR_MESSAGE1));
                this.r = r; this.fi = fi;
                x = r * Math.Cos(fi);
                y = r * Math.Sin(fi);
            }
            catch (PointDException e)
            {

                Debug.WriteLine(e.Message);
                r = -r;
                this.r = r; this.fi = fi;
                x = r * Math.Cos(fi);
                y = r * Math.Sin(fi);
            }
        }

        /// <summary>
        /// Random конструктор
        /// создает точку со случайными вещественными координатами 
        /// в пределах окружности заданного радиуса MaxR
        /// </summary>
        /// <param name="MaxR">радиус окружности</param>
        public PointD(double MaxR)
        {
            //Random rnd = new Random();

            r = rnd.NextDouble() * MaxR;
            fi = rnd.NextDouble() * Math.PI * 2;
            x = r * Math.Cos(fi);
            y = r * Math.Sin(fi);

        }
        /// <summary>
        /// Random конструктор
        /// создает точку со случайными целочисленными
        /// декартовыми координатами 
        /// в пределах квадрата со стороной Max
        /// </summary>
        /// <param name="MaxR">радиус окружности</param>
        public PointD(int Max)
        {
            x = rnd.Next(Max);
            y = rnd.Next(Max);
            r = Math.Sqrt(x * x + y * y);
            fi = Math.Atan2(y, x);
        }
        /// <summary>
        /// Конструктор копии
        /// </summary>
        /// <param name="p">копируемая точка</param>
        public PointD(PointD p)
        {
            this = p;
        }


        //Процедуры-свойства

        /// <summary>
        /// Координата x
        /// </summary>
        public double X
        {
            get { return x; }
        }
        /// Координата y
        /// </summary>
        public double Y
        {
            get { return y; }
        }
        /// Координата r
        /// </summary>
        public double R
        {
            get { return r; }
        }
        /// Координата fi
        /// </summary>
        public double FI
        {
            get { return fi; }
        }

        //переопределение методов родителя класса

        public override bool Equals(object point1)
        {

            if (point1.GetType().Name != "PointD") return false;

            PointD p = (PointD)point1;

            return (Math.Abs(this.x - p.x) < EPS) &&
                (Math.Abs(this.y - p.y) < EPS);
        }
        public override int GetHashCode()
        {
            return Convert.ToInt32(r + fi);
        }
        public override string ToString()
        {
            string s = "x = {0}, y = {1}, r = {2}, fi = {3}";
            s = String.Format(s, x, y, r, fi);
            return s;
        }

        //Методы класса

        /// <summary>
        /// переместить точку по горизонтали
        /// </summary>
        /// <param name="dx">смещение по оси x</param>        
        public void MoveX(double dx)
        {
            x += dx;
            FromDecartToPolar();
        }

        /// <summary>
        /// переместить точку по вертикали
        /// </summary>
        /// <param name="dy">смещение по оси y</param>
        public void MoveY(double dy)
        {
            y += dy;
            FromDecartToPolar();
        }

        /// <summary>
        /// переместить точку по радиусу
        /// </summary>
        /// <param name="dr">Растяжение (dr > 1) или 
        /// сжатие (dr < 1) по радиусу</param>
        public void MoveR(double dr)
        {
            r *= dr;
            FromPolarToDecart();
        }

        /// <summary>
        /// повернуть точку по окружности
        /// </summary>
        /// <param name="dfi">поворот против часовой стрелки</param>
        public void MoveFI(double dfi)
        {
            fi += dfi;
            while (fi > Math.PI * 2) fi -= Math.PI * 2;
            while (fi < -Math.PI * 2) fi += Math.PI * 2;

            FromPolarToDecart();
        }

        //Определение операций класса

        public static bool operator ==(PointD p1, PointD p2)
        {
            return p1.Equals(p2);
        }
        public static bool operator !=(PointD p1, PointD p2)
        {
            return !p1.Equals(p2);
        }

        //Закрытые методы класса

        void FromDecartToPolar()
        {
            r = Math.Sqrt(x * x + y * y);            
            fi = Math.Atan2(y, x);
        }
        void FromPolarToDecart()
        {
            x = r * Math.Cos(fi);
            y = r * Math.Sin(fi);
        }

        //Модуль класса
        static Random rnd;
        /// <summary>
        /// Статический конструктор
        /// </summary>
        static PointD()
        {
            rnd = new Random();
        }
    }

    //Дополняющий класс исключительных ситуаций

    public class PointDException : Exception
    {
        public PointDException()
        { }
        public PointDException(string message)
            : base(message)
        { }
        public PointDException(string message, Exception e)
            : base(message, e)
        { }
    }
}
